-- flateval.e by CChris <oedoc@free.fr>   - Aug. 2007
-- Original code by Martin Stachon.
-- Outputs a sequence of flat tokens rather than nested expressions when tokenizing.
-- Other improvements:
-- * support for several variables and user supplied functions.
-- * better error messages

include get_eval.e
include misc.e -- because of arcsin() etc
include wildcard.e
without warning
with trace

-- numerical constants
constant
    -- for tokenize() function
    T_NUMBER   = 1,  
    T_NAME     = 2

constant
    -- token types
    ERROR            = -1,
    EXPRESSION       =  1,
    OPERATOR         =  2,
    NUMBER           =  3,
    FUNCTION         =  4,
    VARIABLE         =  5,
    END              =  6

constant
    LEFT_RIGHT = 1, -- the function needs left and right expression
    LEFT       = 2, -- -||-               only left      expression
    RIGHT      = 3  -- -||-               only right     expression

-- operator definitions
constant
    OPERATORS = "+-/*^"

-- wrappers for operators
function plus(atom a, atom b)
    return a+b
end function

function minus(atom a, atom b)
    return a-b
end function

function divide(atom a, atom b)
    if b = 0 then return {ERROR,"Attempt to divide by 0."} end if
    return a/b
end function

function multipl(atom a, atom b)
    return a*b
end function

function to_power(atom x, atom n)
    return power(x,n)
end function

constant op_defs = 
{
{'+', LEFT_RIGHT, routine_id("plus")},
{'-', LEFT_RIGHT, routine_id("minus")},
{'/', LEFT_RIGHT, routine_id("divide")},
{'*', LEFT_RIGHT, routine_id("multipl")},
{'^', LEFT_RIGHT, routine_id("to_power")}
}

constant ops_by_precedence = {1,1,2,2,3}

-- wrappers or builtin math functions - some preventive error trapping added
function sinus(atom x)
    return sin(x)
end function

function cosinus(atom x)
    return cos(x)
end function

function abs(atom x)
    if x >= 0 then
        return x
    else
        return -x
    end if
end function

function tangens(atom x)
    if abs(remainder(x-PI/2.0,PI))<1.0e9 then return {ERROR,"Attempt to take the tangent of an odd multiple of PI/2."} end if
    return tan(x)
end function

function logarithm(atom x)
    if x <= 0 then return {ERROR,"Attempt to take the logarithm of a negative number."} end if
    return log(x)
end function

function arcsinus(atom x)
    if x < -1 or x > 1 then return {ERROR,"Attempt to take the arcsinus of a number of magnitude not less than 1."} end if
    return arcsin(x)
end function

function arccosinus(atom x)
    if x < -1 or x > 1 then return {ERROR,"Attempt to take the arccosinus of a number of magnitude not less than 1."} end if
    return arccos(x)
end function

function arctangens(atom x)
    return arctan(x)
end function

function squareroot(atom x)
    if x<0 then return {ERROR,"Attempt to take the square root of a negative number."} end if
    return sqrt(x)
end function

function ran(atom x)
    if not integer(x) or x<=0 then return {ERROR,"Bad argument for ran()."} end if
    return rand(x)
end function

function max(object x, object y)
    if compare(x,y)>=0 then return x
    else return y
    end if
end function

function min(object x, object y)
    if compare(x,y)<=0 then return x
    else return y
    end if
end function

constant func_defs = 
{
{"sin",    routine_id("sinus")},
{"cos",    routine_id("cosinus")},
{"tan",    routine_id("tangens")},
{"tg",     routine_id("tangens")},
{"log",    routine_id("logarithm")},
{"arcsin", routine_id("arcsinus")},
{"arccos", routine_id("arccosinus")},
{"arctan", routine_id("arctangens")},
{"arctg",  routine_id("arctangens")},
{"sqrt",   routine_id("squareroot")},
{"abs",    routine_id("abs")},
{"ran",    routine_id("ran")},
{"rand",   routine_id("ran")},
{"random", routine_id("ran")}
}

-- scanning helpers
function is_digit(integer x)
    if x < '0' then
        return 0
    else
        return x <= '9'
    end if
end function

function is_strict_alpha(integer c)
    if c<'A' or c>'z' then return 0
    elsif c>'Z' and c<'a' then return 0
    else return 1
    end if
end function

function is_alpha(integer c)
    if is_digit(c) or c='_' then return 1
    else return is_strict_alpha(c)
    end if
end function

global function tokenize(object exp)
--/func tokenize(object exp)
--/desc Takes an expression string, and returns a flat sequence of tokens.
--/ret (SEQUENCE) A sequence of flat tokens representing the supplied expression, which can be passed to /eval().
-- The expression string should be such that Euphoria would accept it on the right of an equal sign.
-- Only atomic values are allowed, including as intermediate results inside the expression. /n
-- Known operators are + ,- ,* ,/ and ^ (raise to some power). /n
--
-- Functions can be used, either the ones from the library, which wrap builtins, or ones which
-- are global so as to be visible from the library. Provided finctions are:
--/li /b sin(): the sine function
--/li /b cos(): the cosine function
--/li /b tan(): the tangent function
--/li /b tg():  alias for /b tan
--/li /b log(): the natural logarithm function (argument must be >0)
--/li /b arcsin(): the reverse sine function (argument must be between -1 and 1 included)
--/li /b arccos(): the reverse cosine function (argument must be between -1 and 1 included)
--/li /b arctan(): the reverse tangent function
--/li /b arctg(): alias for /b arctan
--/li /b sqrt(): the square root function (argument must be >=0)
--/li /b abs(): the absolute value finction
--/li /b rand(): the rand() function
--/li /b ran(): alias for /b rand
--/li /b random(): alias for /b rand
--/li /b min(): the minimum of two atoms
--/li /b max(): the maximum of two atoms
--
-- If the expression string is badly formed in any way, a sequence made of the ERROR constant
-- and of a string explaining the problem is returned. Otherwise, a sequence of tokens is returned.
-- The returned value, if valid, ends with an END-type token.
-- This sequence can be manipulated in various ways.
--
-- token formats and types are as follows:
--/li {ERROR, message string}
--/li {EXPRESSION, position of last token}
--/li {OPERATOR,routine_id}
--/li {NUMBER, numeric value as an atom}
--/li {FUNCTION,routine_id, start of enclosing expression, end of first argument, end of second argument, ...}
--/li {VARIABLE, variable index in list of names}
--/li {END,list of variable names, variable position list, {operator position list,operator precedence list,operator start of group list,operator left operand start list},miscellaneous flags, operator precedence}
--
-- Notes: /n
-- if a FUNCTION token has length 4 and its two items are equal, then the argument list is empty. /n
-- in the operator precedence sequence (last item of the last token), 0 means not an operator.
--
-- Example:
--/code
-- sequence s
-- s=tokenize("5*x+1+(3-y/2)*(z+sin(z+2))")
-- ?s   -- screen displays the following:
-- {
--   {3,5},
--   {2,3},
--   {5,1},
--   {2,0},
--   {3,1},
--   {2,0},
--   {1,12},
--   {3,3},
--   {2,1},
--   {5,2},
--   {2,2},
--   {3,2},
--   {2,3},
--   {1,21},
--   {5,3},
--   {2,0},
--   {4,5,18,21},
--   {1,21},
--   {5,3},
--   {2,0},
--   {3,2},
--   {
--     6,
--     {
--       {120},
--       {121},
--       {122}
--     },
--     {
--       {3},
--       {10},
--       {15,19}
--     },
--     {
--       {2,4,6,9,11,13,16,20},
--       {2,1,1,1,2,2,1,1},
--       {0,0,0,7,7,0,14,18},
--       {1,3,5,8,10,7,15,19}
--     },
--     0,
--     {0,2,0,1,0,1,0,0,1,0,2,0,2,0,0,1,0,0,0,1,0}
--   }
-- }
--/endcode

    sequence varnames
    sequence wherevars
    sequence result
    integer last_token
    integer this_char
    sequence new_token
    integer this_function
    integer bracket_level
    integer var_num
    integer pos,op
    sequence last_brackets
    sequence op_prec, op_pos, op_left, op_group
    sequence fun_calls
    integer curr_call
    integer last_closed
    
    if atom(exp) then
        return {NUMBER,exp}
    elsif length(exp)=0 then
        return exp
    end if
    pos=exp[$]
    if pos=26 then
        exp[$]=0
    elsif pos then
        exp &= 0
    end if

    last_token = 0
    result = {}
    new_token = {}
    pos = 0
    varnames = {}
    wherevars = {}
    last_brackets = {}
    bracket_level = 0
    op_prec = {}
    op_pos={}
    op_left={}
    op_group={}
    fun_calls = {}
    curr_call = 0

    while pos < length(exp) do
              trace(1)
        pos +=  1 
        this_char = exp[pos]
        op = find(this_char,OPERATORS)
        if not op then
            last_closed=0
        end if
        if not find(this_char, " \t\r\n") then
            -- skip white spaces
            if this_char = '(' then
                if last_token = T_NAME then -- function call
                                            -- the list of arguments is an expression
                    for f=1 to length(func_defs) do
                        if equal(func_defs[f][1], new_token )  then -- function found
                            result = append(result, {FUNCTION,func_defs[f][2],2+length(result)})
                            this_function = f
                            exit
                        end if
                    end for
                    if not this_function then
                        this_function = routine_id(new_token)
                        if this_function>-1 then
                            result = append(result, {FUNCTION,this_function,1+length(result)})
                        else
                            return {ERROR,"Unrecognised function call."}
                        end if
                    end if
                    curr_call +=1
                    if curr_call > length(fun_calls) then
                        fun_calls &= length(result)
                    else
                        fun_calls[curr_call] = length(result)
                    end if
                    last_token = 0
                elsif last_token = T_NUMBER then -- '(' cannot follow a number
                    return {ERROR,"An opening parenthese cannot follow a number."}
                end if
                -- second element is last token of expression - patched later
                bracket_level +=1
                result = append(result,{EXPRESSION,0})
                -- update list
                if bracket_level <= length(last_brackets) then
                    last_brackets[bracket_level] = length(result)
                else
                    last_brackets &= length(result)
                end if
            elsif last_token=T_NAME then
                var_num = find(new_token,varnames)
                result = append(result,{VARIABLE,var_num})
                if not var_num then
                    varnames = append(varnames,new_token)
                    result[$][2] = length(varnames)
                    wherevars = append(wherevars,{length(result)})
                else
                    wherevars[var_num] &= length(result)
                end if
            end if          
            if this_char=0 then
                exit
            elsif this_char = ')' then 
                if bracket_level <= 0 then
                    return {ERROR,"Extra closing parenthese."}
                end if
                this_function = last_brackets[bracket_level]
                last_closed = this_function
                if this_function > 1 and result[this_function-1][1]=FUNCTION then
                    last_closed-=1
                end if
                result[this_function][2] = length(result)
                if curr_call and this_function = fun_calls[curr_call]+1 then -- closing a call
                    result[this_function-1] &= length(result)                -- end of last argument
                    curr_call -= 1
                elsif this_function = length(result) then -- empty pair, not expected
                    return {ERROR,"Empty parentheses pairs are not valid, unless they follow a function call."}
                end if
                bracket_level -=1
                last_token = 0
            elsif op then -- operator found, end token
                if not length(result) or result[$][1] = EXPRESSION or result[$][1]=OPERATOR then
                    return {ERROR,"An operator was found, but was not expected."}
                else
                    result = append(result,{OPERATOR, op_defs[op][3]})
                end if
                last_token = 0
            elsif is_digit(this_char) then -- new token found
                new_token = valuex(exp[pos..$])
                if new_token[1]!=GET_SUCCESS then
                    return  {ERROR,"Badly formed number in expression."}
                end if
                new_token[1]=NUMBER
                pos+=new_token[3]-1
                result=append(result,new_token[1..2])
                last_token = T_NUMBER
            elsif is_strict_alpha(this_char) then -- new token found
                for i=pos+1 to length(exp) do
                    if not is_alpha(exp[i]) then
                        new_token=exp[pos..i-1]
                        pos=i-1
                        exit
                    end if
                end for
                last_token = T_NAME
            elsif this_char=',' then
                if curr_call<=0 then
                    return {ERROR,"A comma was found outside a function call."}
                end if
                last_token = 0
                result[fun_calls[curr_call]] &= length(result)
            elsif this_char = '(' then -- done already
                last_token = 0
            else
                return {ERROR,"Illegal character in expression."}
            end if
        end if
        if op then
            op_prec &= ops_by_precedence[op]
            op_pos &= length(result)
            if last_closed then
                op_left &= last_closed
            else
                op_left &= length(result)-1
            end if
            if bracket_level then
                this_char = last_brackets[bracket_level]
            else
                this_char = 0
            end if
            if curr_call then
                this_function = result[fun_calls[curr_call]][$]
            else
                this_function = 0    
            end if
            op_group &= max(this_function,this_char)
        end if
    end while -- reached end of input
    if bracket_level then
        return {ERROR,"Too many open parentheses."}
    end if
    -- there are no unfinished tokens
    wherevars = {END,varnames,wherevars,{op_pos,op_prec,op_group,op_left},0,0}
    varnames = repeat(0,length(result))
    for i=1 to length(op_prec) do
        varnames[op_pos[i]] = op_prec[i]
    end for
    wherevars[$] = varnames
    return append(result,wherevars)
end function

sequence toks,vnames,oprecs,active

constant IS_NORMALISED=1, IS_COALESCED=2

function pos_max(sequence s)
    integer n,p

    p=s[1]
    n=1
    for i=2 to length(s) do
        if s[i]>p then
            n=i
            p=s[i]
        end if
    end for
    if s[n] then 
        return n 
    else 
        return 0 
    end if
end function

function next_active(integer n)
    for i=n to length(toks) do
        if active[i]=1 then
            return i
        end if
    end for
    return 0
end function

function prev_active(integer n)
    for i=n to 1 by -1 do
        if active[i]=1 then
            return i
        end if
    end for
    return 0
end function

integer r_eval_
function eval_func(integer p,integer q) -- assumes toks[p][1]=FUNCTION
    sequence args
    object left
    integer n

    if toks[p+1][1]!=EXPRESSION then
        return {ERROR,"A function call must be followed by its argument list."}
    end if
    if toks[p][3] = toks[p][4] then -- no args
        args = {}
    else
        args = repeat(0,length(toks[p])-3)
        for i=1 to length(args) do
            left = call_func(r_eval_,{toks[p][i+2]+1,toks[p][i+3]})
            if sequence(left) then
                return {ERROR,"Error in evalutaing a function argument: "&left[2]}
            end if
            args[i] = left
        end for
    end if
    n = toks[p+1][2]
    left = call_func(toks[p][2],args)
    if sequence(left) then
        return {ERROR,"Error in function evaluation: "&left[2]}
    end if
    toks[n] = {NUMBER,left}
    active[p..n-1]=0
    active[n]=1
    return call_func(r_eval_,{n,q})
end function

function eval_expr(integer p) -- assumes toks[p][1]=EXPRESSION
    integer n
    object left

    n = toks[p][2]
    left = call_func(r_eval_,{p+1,n})
    if sequence(left) then 
        return {ERROR,"Error evaluating an expression "&left[2]}
    end if
    active[p..n-1]=0
    active[n]=1
    toks[n]={NUMBER,left}
    return n
end function

function eval_(integer p,integer q)
    object left
    integer n,p1,nbef,naft

    p = next_active(p)
    if not p then
        return {ERROR,"Internal error: nothing to evaluate."}
    end if
    while p<q and not active[q] do
        q-=1
    end while
    if atom(toks[p]) then
        if toks[p] = NUMBER then
            return toks[p+1]
        else
            return {ERROR,"Badly formed token sequence."}
        end if
    elsif sequence(toks[p][1]) then
        toks[p] = toks[p][1]
    end if
    if p=q and toks[p][1]=NUMBER then
        return toks[p][2]
    elsif toks[p][1]=EXPRESSION then
        if toks[p][2]=q then
            return eval_(p+1,q)
        else 
            return eval_(eval_expr(p),q)
        end if
    elsif toks[p][1]=FUNCTION then    
        return eval_func(p,q)
    else
        p1 = p
        while p1 > 0 and p1 <= q do  
            if toks[p][1]=ERROR then
                return {ERROR,"Badly formed token sequence."}
            elsif toks[p][1]=VARIABLE  then
                return {ERROR,"Unevaluated variable."}
            elsif toks[p1][1]=FUNCTION then
                if sequence(eval_func(p1,q)) then
                    return {ERROR,"Error in evaluating a function."}
                end if
                exit
            elsif toks[p1][1]=EXPRESSION then
                p1 = eval_expr(p1)
            else
                p1 = next_active(p1+1)
            end if
        end while
        -- no more parentheses or function calls
        while 1 do
            n = pos_max(oprecs[p..q]*active[p..q])+p-1
            if n>=p then -- assumes all operators are LEFT_RIGHT
                nbef=prev_active(n-1)
                naft=next_active(n+1)
                toks[naft]={NUMBER,call_func(toks[n][2],{toks[nbef][2],toks[naft][2]})}
                active[nbef..naft-1]=0
            else
                p=find(1,active[p..q])+p-1  -- must be a number
                active[p]=0
                return toks[p][2]
            end if
        end while
    end if
end function
r_eval_ = routine_id("eval_")

global function eval( object exp, sequence values )  -- exp is tokenized
--/func eval( object exp, sequence values )
--/desc Attempts to evaluate a valid sequence of tokens, given some variable values.
--/ret (OBJECT) A numerical value on success, a {ERROR,message string} pair else.
-- /i exp is a sequence of tokens, presumably obtained from /i tokenize(), even though any properly formed sequence will do.
-- /i values is a list of pairs {variable name, value}. Only ayoms are allowable values.
-- The list may terminate with a single atom, in which case this becomes the default value for variables which do not appear in /i values but do inside /i exp.
-- If not default value is supplied, 0.0 will be used.
--
-- If any error occurs, either because of invalid values or corrupted daa, then a sequence is returned, the second element of which summarily explains the problem.
--
-- Example:
--/code
-- sequence s
-- s=tokenize("5*x+1+(3-y/2)*(z+sin(z+2))")
-- ?eval(s,{{"x",1},{"y",2},{"z",3}})  -- prints 10.08215145
-- ?machine_func(26,0)
    sequence vvalues,element,varpos
    atom def_value
    integer n
    
    if compare(exp,{})<1 then
        return exp
    end if
    varpos = exp[$]
    if varpos[1]!=END then
        return {ERROR,"Missing END token at the end of expression - most likely currupted."}
    else       
        toks = exp
        vnames = varpos[2]
        oprecs = varpos[$]
        n = length(values)
        if sequence(values[$]) then
            def_value = 0.0
        else
            def_value = values[$]
            n-=1
        end if
        vvalues = repeat(def_value, length(vnames))
        for i=1 to n do
            element=values[i]
            if length(element)=0 or atom(element[1]) then
                return {ERROR,"An assignment pair must start with a variable name."}
            elsif length(element)=1 then
                element &= def_value
            elsif length(element)>2 then
                element=element[1..2]
            end if
            if sequence(element[2]) then
                return {ERROR,"Values of variables must be atoms."}
            end if
            n = find(element[1],vnames)
            if n then
                vvalues[n] = element[2]
            end if
        end for
        active = {NUMBER,0}
        varpos = varpos[3]
        for i=1 to length(vnames) do -- evaluate variables
            element = varpos[i]
            active[2] = vvalues[i]
            for j=1 to length(element) do
                toks[element[j]] = active
            end for
        end for
        active = repeat(1,length(toks)-1)&0 -- the end token is never active
        return eval_(1,length(toks))
    end if
end function   

function insert_expr(sequence s,integer opmin,integer op_max)
    sequence token,last_s
    integer p

    for i=1 to length(s)-1 do
        token=s[i]
        if token[1]=EXPRESSION then
            p=s[i][2]
            s[i][2]+=(p>opmin)
        elsif token[1]=FUNCTION then
            token=token[3..$]
            s[i][3..$]+=(token>opmin)
        end if
    end for
    last_s=s[$]
    token=last_s[3]
    s[$][3]+=(token>=opmin)
    token=last_s[$]
    token=token[1..opmin-1] & 0 & token[opmin..$]
    s[$][$]=token
    return append(s[1..opmin-1],{EXPRESSION,op_max+1}) & s[opmin..$]
end function

global function normalize(sequence s)
--/func normalize(sequence s)
--/desc Adds parenthesis so that each operator is alone inside a parenthese pair or function argulment.
--/ret (SEQUENCE) Another tokenised sequence.
-- /i s is tokenised, and was presumaly obtained through a call to /tokenize().
 Example:
--/code
-- sequence s
-- s=tokenize("5*x+1+(3-y/2)*(z+sin(z+2))")
-- s=normalize(s)
-- ?s   -- screen displays the following:
-- {
--   {1,7},
--   {1,5},
--   {3,5},
--   {2,3},
--   {5,1},
--   {2,0},
--   {3,1},
--   {2,0},
--   {1,25},
--   {1,16},
--   {3,3},
--   {2,1},
--   {1,16},
--   {5,2},
--   {2,2},
--   {3,2},
--   {2,3},
--   {1,25},
--   {5,3},
--   {2,0},
--   {4,5,22,25},
--   {1,25},
--   {5,3},
--   {2,0},
--   {3,2},
--   {
--     6,
--     {
--       {120},
--       {121},
--       {122}
--     },
--     {
--       {5},
--       {14},
--       {19,23}
--     },
--     {
--       {2,4,6,9,11,13,16,20},
--       {2,1,1,1,2,2,1,1},
--       {0,0,0,7,7,0,14,18},
--       {1,3,5,8,10,7,15,19}
--     },
--     1,
--     {0,0,0,2,0,1,0,1,0,0,0,1,0,0,2,0,2,0,0,1,0,0,0,1,0}
--   }
-- }
--/endcode
    sequence x,oprecs,oppos,opgrp,opleft,sl_precs,sl_index
    integer p,q,q1,op_min,op_max

    if and_bits(s[$][5],IS_NORMALISED) then return s end if -- expression already normalised
    x=s[$][4]
    oppos=x[1]
    oprecs=x[2]
    opgrp=x[3]
    opleft=x[4]
    for i=length(opleft) to 2 by -1 do 
        p=opgrp[i]
        q=find(p,opgrp)
        if q<i then -- two or more ops in the same group
                    -- all those may not be contiguous, as in 1+(1+1)+1
            sl_precs={oprecs[q]}
            sl_index={q}
            for j=q+1 to i do
                if opgrp[j]=p then
                    sl_precs &= oprecs[j]
                    sl_index &= j
                end if
            end for
            for u=1 to length(sl_index)-1 do
                q1=pos_max(sl_precs)
                if q1=0 then exit end if
                sl_precs[q1]=0
                q=sl_index[q1]
                -- locate boundaries
                op_min=opleft[q] -- first token of left group
                op_max=oppos[q]+1 -- last token
                if s[op_max][1]=EXPRESSION then
                    op_max=s[op_max][2]
                elsif s[op_max][1]=FUNCTION then
                    op_max=s[op_max+1][2]
                end if
                opgrp+=(opgrp>=op_min)
                opgrp[q]=op_min
                oppos+=(oppos>=op_min)
                opleft+=(opleft>=op_min)
                s=insert_expr(s,op_min,op_max)
                q1=find(oppos[q]+1,opleft)
                if q1 then
                    opleft[q1]=op_min
                end if   
            end for
        end if
    end for
    s[$][5]+=IS_NORMALISED
    return s
end function


